class Goisi <QRCanvasRectangle
   attr_reader :color
   def initialize(canvas,color)
      @color=color
      super(canvas)
      if color=='B'
         @pix = canvas.pixB
      else
         @pix = canvas.pixW
      end
      setSize(@pix.width, @pix.height)
      @dx=@dy=-(@pix.width-1)/2
   end
 
   def drawShape(painter)
      painter.drawPixmap(x, y, @pix)
   end

   def move(x,y)
      super(x+@dx,y+@dy)
   end
end

class Cursor <QRCanvasRectangle
   attr_reader :xx
   attr_reader :yy

   def initialize(canvas)
      super(canvas)
      changeSize(canvas.sz)
      setZ(1)
   end
 
   def changeSize(sz)
      @sz=sz
      f = "#{$PIXDIR}cursor#{sz}.png"
      f = "#{$PIXDIR}cursor19.png" if sz==15||size==14
      @pix=QPixmap.new(f)
      setSize(@pix.width, @pix.height)
      @dx=@dy=-(@pix.width-1)/2

      @xx=@yy=(sz+1)/2
   end

   def drawShape(painter)
      painter.drawPixmap(self.x, self.y, @pix)
   end

   def move(x,y)
      super(x+@dx,y+@dy)
   end

   def moveTo(x,y=nil)
      if y==nil
         case x
         when :right
            @xx+=1
            @xx=1 if @xx>@sz
         when :left
            @xx-=1
            @xx=@sz if @xx<=0
         when :down
            @yy+=1
            @yy=1 if @yy>@sz
         when :up
            @yy-=1
            @yy=@sz if @yy<=0
         when :center
            @xx=@yy=(@sz+1)/2
         end
      else
         @xx=x
         @yy=y
      end
   end
end

class GobanText < QCanvasText
   @@black = QColor.new(0,0,0)
   @@white = QColor.new(255,255,255)
   @@red   = QColor.new(255,0,0)
   def initialize(s,canvas)
      super(s,canvas)
      @dx=0
      @dy=0
      setTextFlags(Qt::AlignLeft|Qt::AlignTop)
   end
   def setFont(f)
      super
      r=boundingRect
      @dx = -r.width*0.5
      @dy = -r.height*0.5
   end
   def move(x,y)
      super(x+1+@dx,y+1+@dy)#Ĵ
   end
end

class Number < GobanText
   @@font = Hash.new
   @@font[19] = QFont.new('lcfont',14)
   @@font[13] = QFont.new('lcfont',18)
   @@font[9]  = QFont.new('lcfont',24)

   def initialize(canvas,color,n)
      super(n.to_s,canvas)
      setFont(@@font[canvas.sz])
      setColor(@@white) if color=='B'
      setColor(@@black) if color=='W'
      setZ(2)
      @color=color
   end

   def changeColor
      if @color=='B'
         setColor(@@black)
         @color='W'
      else
         setColor(@@white)
         @color='B'
      end
   end

   def iro
      @color
   end
end

class Label < GobanText
   @@font16=QFont.new('lcfont',16)
   @@font18=QFont.new('lcfont',18)
   @@font20=QFont.new('lcfont',20)
   @@font24=QFont.new('lcfont',24)

   def initialize(canvas,s)
      case canvas.sz
      when 19
         if s.length>1# 
            f=@@font16
         else
            f=@@font18
         end
      when 13
         f=@@font20
      when  9
         f=@@font24
      end
      super(EUCstr.new(s).toUTF,canvas)
      setFont(f)

      setColor(@@red)
      setZ(3)
   end
end

class DeadStoneMark < Label 
   def initialize(canvas)
      super(canvas,'')
      setZ(4)
   end
end

class GobanCanvas < QCanvas
   MKchar = { 'CR' => '','MA' => '','M'  => '' ,\
              'SQ' => '','TR' => '','SL' => ''  }
   attr_reader :pixB
   attr_reader :pixW
   attr_reader :sz
   attr_reader :pointedAt

   def initialize
      super(404,404)
      @pointedAt = RSignal.new('Object')
      @sz=19
      @goisi=Array2.new(@sz)
      @number=Array.new
      @label=Array.new
      @cursor=Cursor.new(self)
      changeSize(@sz)
      @optSave=nil
   end

   def changeSize(sz)
      clear
      @sz=9
      @sz=13 if sz>9
      @sz=19 if sz>13
      @currentBoard=Board.new(@sz)

      fb = "#{$PIXDIR}kurostone#{@sz}.png"
      fw = "#{$PIXDIR}sirostone#{@sz}.png"
      fg = "#{$PIXDIR}goban#{@sz}.png"
      fg = "#{$PIXDIR}goban15.png" if sz==15 ||sz==14
      @pixB=QPixmap.new(fb)
      @pixW=QPixmap.new(fw)
      setBackgroundPixmap( QPixmap.new(fg) )

      @goisiHalfWidth=(@pixB.width-1)/2
      @width1ro=@pixB.width+1
      case @sz
      when 19
         @coor11=30
      when 13
         @coor11=36
         @width1ro=@pixB.width
      when  9
         @coor11=42
      end

      @deadstoneMark && @deadstoneMark.clear
      @moveStoneMark=nil

      @cursor.changeSize(sz)
      moveCursor(:center)
   end

   def clear
      @goisi.each_index{|i|
         if @goisi[i].is_a?(Goisi)
            @goisi[i].setVisible(false)
            @goisi[i]=nil
         end
      }
      @number.each_index{|i|
         if @number[i].is_a?(Number)
            @number[i].setVisible(false)
            @number[i]=nil
         end
      }
      @number.clear
      GC.start
      a=allItems()
      for i in 0..a.count-1
         removeItem(a.at(i))
      end
      @deadstoneMark && @deadstoneMark.clear
      @moveStoneMark=nil
      @currentBoard=Board.new(@sz)
   end

   def slotTapped(x,y)
      x=x-@coor11+@goisiHalfWidth
      y=y-@coor11+@goisiHalfWidth
      return if x%@width1ro==@width1ro-1
      return if y%@width1ro==@width1ro-1
      x=x/@width1ro+1
      y=y/@width1ro+1
      return if x<=0 || x>@sz
      return if y<=0 || y>@sz

      if @useCursor && !(x==@cursor.xx && y==@cursor.yy)
         moveCursor(x,y)
         update
      else
         @pointedAt.sendWith([x,y])
      end
   end

   def draw(node,option)
      board=node.board
      opt=option['number']
      if opt!=@optSave
         if opt==true
            for y in 1..board.size
            for x in 1..board.size
               n=@currentBoard[x,y].abs
               if n>0 && n!=Board::BlackInitial
                  if @number[n]!=nil
                     @number[n].setVisible(true)
                  end
               end
            end
            end
         else
            for i in 1..@currentBoard.n-1
               if @number[i]!=nil
                  @number[i].setVisible(false)
               end
            end
            if opt==false
               n=@currentBoard.n
               if n>0 && @number[n]!=nil
                  @number[n].setVisible(false)
               end
            end
         end
      end
      @optSave=opt

      if opt==true
         for y in 1..board.size
         for x in 1..board.size
            next if @currentBoard[x,y]==board[x,y]
            next if @currentBoard[x,y]==Board::NoStone

            n=@currentBoard[x,y].abs
            if n!=Board::BlackInitial
               @number[n].setVisible(false)
            end
         end
         end
      elsif opt!=false
         n=@currentBoard.n
         @number[n].setVisible(false) if @number[n]!=nil
      end

      for y in 1..board.size
      for x in 1..board.size
         if board[x,y]==@currentBoard[x,y]
            if opt!=true && opt!=false
               n=board[x,y].abs
               if n>0 && n==board.n
                  @number[n].setVisible(true)
               end
            end
            next
         end

         if board[x,y]==Board::NoStone
            @goisi[x,y].setVisible(false)
         else
            xx,yy=calcCoor(x,y)
            c='B'
            c='W' if board[x,y]<0
            if @goisi[x,y]!=nil && @goisi[x,y].color==c
               @goisi[x,y].setVisible(true)
            else
               if @goisi[x,y]!=nil
                  @goisi[x,y].setVisible(false)
               end
               @goisi[x,y]=Goisi.new(self,c)
               @goisi[x,y].move(xx,yy)
               @goisi[x,y].show
            end

            n=board[x,y].abs
            next if n==Board::BlackInitial
            if @number[n]==nil
               @number[n]=Number.new(self,c,n)
               @number[n].show
               @number[n].setVisible(opt==true)
            elsif c!=@number[n].iro
               @number[n].changeColor
            end
            @number[n].move(xx,yy)
            if opt==true
               @number[n].setVisible(true)
            elsif opt==false
            else
               if n==board.n
                  @number[n].setVisible(true)
               elsif n>board.n
                  @number[n].setVisible(false)
               end
            end
         end
      end
      end
      @currentBoard=board

      @label.each{|l|
         l.setVisible(false)
      }
      @label.clear
      if option['mark']
         ['LB','CR','MA','M','SL','SQ','TR'].each{|pn|
            if node.propList[pn]!=nil
               node.propList[pn].each{|pv|
                  if pn=='LB'
                     next if !(pv =~ /^[a-s][a-s]\:(.)/e)
                     s=$1
                  else  
                     next if !(pv =~ /^[a-s][a-s]$/)
                     s=MKchar[pn]
                  end
                  x=pv[0]-?a+1
                  y=pv[1]-?a+1
                  if x<=@sz && y<=@sz
                     l=Label.new(self,s)
                     xx,yy=calcCoor(x,y)
                     l.move(xx,yy)
                     l.show
                     @label.push(l)
                  end
               }
            end
         }
      end


      update
      GC.start
   end

   def drawDeadstone(list)
      @deadstoneMark || @deadstoneMark=Hash.new
      @deadstoneMark.each_pair{|xy,mark|
         mark.setVisible(list.include?(xy))
      }
      list.each{|xy|
         if @deadstoneMark[xy]==nil
            @deadstoneMark[xy]=DeadStoneMark.new(self)
            xx,yy=calcCoor(xy[0],xy[1])
            @deadstoneMark[xy].move(xx,yy)
            @deadstoneMark[xy].show
         end
      }
      update
   end

   def drawMoveStoneMark(xy)
      if xy==nil
         @moveStoneMark.setVisible(false)
      else
         if @moveStoneMark==nil
            @moveStoneMark=Label.new(self,'')
         end
         xx,yy=calcCoor(xy[0],xy[1])
         @moveStoneMark.move(xx,yy)
         @moveStoneMark.setVisible(true)
         @moveStoneMark.show
      end
      update
   end


   def calcCoor(x,y)
      xx=(x-1)*@width1ro+@coor11
      yy=(y-1)*@width1ro+@coor11
      return xx,yy
   end

   def setCursorOn(f)
      @useCursor=f
      if f
         @cursor.show
         moveCursor(:center)
      else
         @cursor.hide
      end
      update
   end
   def isCursorOn
      @useCursor
   end
   def moveCursor(x,y=nil)
      @cursor.moveTo(x,y)
      xx,yy=calcCoor(@cursor.xx,@cursor.yy)
      @cursor.move(xx,yy)
      update
   end
   def okPressed
      @pointedAt.sendWith([@cursor.xx,@cursor.yy])
   end
end

class GobanView <QRCanvasView
   def initialize(canvas, parent = nil, name = '')
      super(canvas, parent, name)
      @canvas=canvas
      setFixedSize(canvas.width+4,canvas.height+4)
      setFocusPolicy(QWidget::ClickFocus)

      @cursorPressed = RSignal.new('int')
      @alphabetPressed = RSignal.new('int')
      @shiftAlphabetPressed = RSignal.new('int')
      @backSpacePressed=RSignal.new
      catchEvent

      @shiftPressed=false
   end
   attr_reader :cursorPressed
   attr_reader :alphabetPressed
   attr_reader :shiftAlphabetPressed
   attr_reader :backSpacePressed
 
   def contentsMousePressEvent(ev)
      x=contentsX + ev.x
      y=contentsY + ev.y
      @canvas.slotTapped(x,y)
      self.setFocus
   end

   def keyPressEvent(ev)
      n=0
      case ev.key
      when 4128 #shift
         @shiftPressed=true
      when 4116 #
         if @canvas.isCursorOn && !@shiftPressed 
            @canvas.moveCursor(:right)
         else
            if @shiftPressed && !@canvas.isCursorOn
               n=10
            else
               n=1
            end
         end
      when 46 #륭ݸ
         n=1
      when 4114 #
         if @canvas.isCursorOn && !@shiftPressed 
            @canvas.moveCursor(:left)
         else
            if @shiftPressed && !@canvas.isCursorOn
               n=-10
            else
               n=-1
            end
         end
      when 4115 #
         if @canvas.isCursorOn && !@shiftPressed 
            @canvas.moveCursor(:up)
         else
            if @shiftPressed
               n=-1000
            else
               n=-1
            end
         end
      when 4117 #
         if @canvas.isCursorOn && !@shiftPressed 
            @canvas.moveCursor(:down)
         else
            if @shiftPressed
               n=1000
            else
               n=1
            end
         end
      when 4099 #BS
         @backSpacePressed.send
      when 4100 #OK return
         if @canvas.isCursorOn
            @canvas.okPressed
         end
      when ?A .. ?Z #a to z
         if @shiftPressed
            if ev.key == ?C
               @shiftPressed=false
            end
            @shiftAlphabetPressed.sendWith(ev.key)
         else
            @alphabetPressed.sendWith(ev.key-?A)
         end
      else
         print 'shift +' if @shiftPressed
         p ev.key
         ev.ignore
      end
      
      @cursorPressed.sendWith(n) if n!=0
   end

   def keyReleaseEvent(ev)
      if ev.key==4128
         @shiftPressed=false
      end
   end
end
