class UndoStack < Array
   MaxSize=10

   def initialize
      @redoStack=Array.new
   end

   def clear
      super
      @redoStack.clear
   end

   def redoAvailable?
      !@redoStack.empty?
   end

   def pushStack(a)
      push(a)
      shift if size>MaxSize
      @redoStack.clear
   end

   def undo(tree)
      return tree if size==0
      @redoStack.push(tree)
      pop
   end

   def redo(tree)
      return tree if @redoStack.size==0
      push(tree)
      @redoStack.pop
   end
end

class TextArea < QMultiLineEdit
   def initialize(par)
      super(par)
      setWordWrap(1)   #WidgetWidth
      setWrapPolicy(1) #Anywhere
      setFont(QFont.new('lcfont',18))
   end
   def setText(s)
      super(s)
      super(s)#for wordwrap
      setCursorPosition(0,0)
   end
   def isSGF?
      s=EUCstr.new(text)
      if text.to_str!=''
         if s =~ /^\s*\(\s*\;/e
            return true
         end
      end
      return false
   end
end

class MoveButtons < QHBox
   def initialize(par)
      super(par)
      setMargin(1)
      setSpacing(1)

      @button=Hash.new
      s=['<-','<<','<','>','>>','->']
      s.each{|str|
         @button[str]=QPushButton.new(str,self,'')
      }
   end
   attr_reader :button

   def draw(node)
      f=(node.parent!=nil)
      if f!=@button['<'].isEnabled
         for s in ['<-','<<','<']
            @button[s].setEnabled(f)
         end
      end
      f=!(node.child.empty?)
      if f!=@button['>'].isEnabled
         for s in ['->','>>','>']
            @button[s].setEnabled(f)
         end
      end
   end
end 

class MoveView < QLabel
   def initialize(par)
      super(par)
      setMinimumSize(100,35)
   end
   def draw(node)
      s=''
      if node.board.is_a?(Board)
         s=sprintf("%3d : ",node.board.n)
         if m=node.move
            s+='' if m.color=='B'
            s+='' if m.color=='W'
            if m.x==0
               s+='ѥ'
            else
               x=m.x
               x+=1 if x>=9
               s+=' '+(x-1+?A).chr
               s+=(node.board.size-m.y+1).to_s
            end
            s+=' '+m.toSGF
         end
         s=EUCstr.new(s).toUTF
      end
      setText(s)
   end
end

class VariationView < QHBox
   def initialize(par)
      super(par)
      @label=QLabel.new(self)
      @select=QComboBox.new(self)
      @select.setFixedSize(60,40)

   end
   attr_reader :label
   attr_reader :select

   def draw(tree)
      @select.clear if @select.count!=0
      @label.clear  if @label.text!=''
      iPar=tree.currentNode.parent
      if iPar!=nil
         if tree[iPar].child.size>1
            s=' Ѳޤ(A'
            s << (tree[iPar].child.size-1+?A).chr
            s << ')'
            @label.setText(EUCstr.new(s).toUTF)
            iSister=0
            tree[iPar].child.each_index{|i|
               iSister=i if tree[iPar].child[i]==tree.currentIndex
               @select.insertItem((i+?A).chr)
            }
            @select.setCurrentItem(iSister)
            setEnabled(true)
         else
            setEnabled(false) if isEnabled
         end
      end
   end
end

class MyMainWindow < QMainWindow
   Normal=0
   GoBackToStone=1
   InitialBoard=2
   PutMark=3
   CountArea=4
   MoveStone=5

   def initialize
      super

      hbox=QHBox.new(self)
      hbox.setSpacing(0)
      setCentralWidget(hbox)

      @ctrlViewStack=QWidgetStack.new(hbox,'')
      @ctrlView=QVBox.new
      @initialBoardView=nil

      idd=@ctrlView.id
      @ctrlViewStack.addWidget(@ctrlView,idd)
      @ctrlViewStack.raiseWidget(@ctrlView)

      setupMenu(@ctrlView)
      @moveView=MoveView.new(@ctrlView)
      @txtArea =TextArea.new(@ctrlView)
      connect(@txtArea, QSIGNAL("textChanged()"), self, 'slotTextChange()')

      @variationView = VariationView.new(@ctrlView)
      Qte::connect(@variationView.select,QSIGNAL("activated(int)"), self, 'gotoVariation()')

      @buttonBox=MoveButtons.new(@ctrlView)
      connect(@buttonBox.button['<-'], QSIGNAL("clicked()"), self, 'initialMove()')
      connect(@buttonBox.button['<<'], QSIGNAL("clicked()"), self, 'prev10Move()')
      connect(@buttonBox.button['<'] , QSIGNAL("clicked()"), self, 'prevMove()')
      connect(@buttonBox.button['>'] , QSIGNAL("clicked()"), self, 'nextMove()')
      connect(@buttonBox.button['>>'], QSIGNAL("clicked()"), self, 'next10Move()')
      connect(@buttonBox.button['->'], QSIGNAL("clicked()"), self, 'lastMove()')

      @goban=GobanCanvas.new
      @gobanView = GobanView.new(@goban, hbox)
      @gobanView.update()
      connect( @gobanView.cursorPressed, self, 'slotCursorPressed' )
      connect( @gobanView.alphabetPressed, self, 'slotAlphabetPressed' )
      connect( @gobanView.shiftAlphabetPressed, self, 'slotShiftAlphabetPressed' )
      connect( @gobanView.backSpacePressed, self,'slotBSpressed')
      connect( @goban.pointedAt, self, 'boardPointedAt')

      @status=Normal
      @fileName=nil
      @fileWritable=false

      @option=Hash.new
      @option['comment']=true
      @option['mark']=true
      @option['number']=3
      setCursorOff

      @acTab=QAccel.new(self) 
      @acTab.insertItem(4097,4097)#tab
      @acTab.connectItem(4097,self,"slotTabPressed()")

      @undo=UndoStack.new
      @tree=nil
      newTree
      setCaption()
      catchEvent
   end

   def setCaption()
      s='ӡθ - '
      if @fileName
         f=File.basename(@fileName)
      else
         f='unnamed'
      end
      s=EUCstr.new(s+f).toUTF     
      super(s)
   end

   def setupMenu(parent)
      @menubar = QPEMenuBar.new(parent)
      @menubar.setSeparator(QMenuBar::InWindowsStyle)
      font = QFont.new('lcfont',18)
      @menubar.setFont(font)

      @menu=Hash.new
      menuTitle=['File','Edit','View','Move','Opt']
      for s in menuTitle
         @menu[s] = QPopupMenu.new
         @menubar.insertItem(s,@menu[s])
      end
      @menuid=Hash.new

      @menu['new']=QPopupMenu.new
      @menu['new'].insertItem("19*19",self,"newTree")
      @menu['new'].insertItem("15*15",self,"newTree15")
      @menu['new'].insertItem("13*13",self,"newTree13")
      @menu['new'].insertItem(" 9* 9",self,"newTree9")
      @menuid['new']=@menu['File'].insertItem("New",@menu['new'])


      @menu['File'].insertSeparator
      @menuid['open']=@menu['File'].insertItem("Open",self,"load()")
      @menuid['loadFromTextArea']=@menu['File'].insertItem("Load SGF from Textarea",self,"loadFromTextArea()")
      @menu['File'].insertSeparator
      @menuid['save']=@menu['File'].insertItem('Save',self,'save()')
      @menuid['saveAs']=@menu['File'].insertItem("Save As",self,'saveAs()')
      @menu['File'].insertSeparator
      @menuid['quit']=@menu['File'].insertItem('Quit', QApplication::qApp,'closeAllWindows')

      s=EUCstr.new('ѥ(P)').toUTF
      @menuid['pass']=@menu['Edit'].insertItem(s,self,"pass")
      @menu['Edit'].insertSeparator
      s=EUCstr.new('ߤμ(ʹ)(&D)').toUTF
      @menuid['removeNode']=@menu['Edit'].insertItem(s,self,"removeNode")
      s=EUCstr.new('ư').toUTF
      @menuid['moveStone']=@menu['Edit'].insertItem(s,self,"moveStone")
      @menu['kakuage']=QPopupMenu.new
      s=EUCstr.new('ѲA˳ʾ夲').toUTF
      @menuid['kakuageToA']=@menu['kakuage'].insertItem(s,self,"kakuageToA")
      s=EUCstr.new('ҤȤĳʾ夲').toUTF
      @menuid['kakuage1']=@menu['kakuage'].insertItem(s,self,"kakuage1")
      s=EUCstr.new('ҤȤĳʲ').toUTF
      @menuid['kakusage1']=@menu['kakuage'].insertItem(s,self,"kakusage1")
      s=EUCstr.new('Ǹ˳ʲ').toUTF
      @menuid['kakusageToZ']=@menu['kakuage'].insertItem(s,self,"kakusageToZ")
      s=EUCstr.new('Ѳ').toUTF
      @menuid['kakuage']=@menu['Edit'].insertItem(s,@menu['kakuage'])

      @menu['Edit'].insertSeparator
      s=EUCstr.new('').toUTF
      @menuid['initialBoard']=@menu['Edit'].insertItem(s,self,'initialBoard')
      s=EUCstr.new('̤̤ˤ').toUTF
      @menuid['current2Initial']=@menu['Edit'].insertItem(s,self,'current2Initial')
      @menu['Edit'].insertSeparator
      s=EUCstr.new('ɲáԽ(&C)').toUTF
      @menuid['addComment']=@menu['Edit'].insertItem(s,self,"addComment")
      s=EUCstr.new('').toUTF
      @menuid['putMark']=@menu['Edit'].insertItem(s,self,'putMark')
      @menu['Edit'].insertSeparator
      s=EUCstr.new('').toUTF
      @menuid['editGameinfo']=@menu['Edit'].insertItem(s,self,'editGameinfo')
      @menu['Edit'].insertSeparator
      @menuid['undo']=@menu['Edit'].insertItem('undo(&U)',self,'undo')
      @menuid['redo']=@menu['Edit'].insertItem('redo(&R)',self,'redo')


      s=EUCstr.new('ϥ').toUTF
      @menuid['viewHama']=@menu['View'].insertItem(s,self,'viewHama')
      s=EUCstr.new('Ϸ׻').toUTF
      @menuid['textGoban']=@menu['View'].insertItem(s,self,'countArea')
      @menu['View'].insertSeparator
      s=EUCstr.new('').toUTF
      @menuid['viewGameinfo']=@menu['View'].insertItem(s,self,'viewGameinfo')
      @menu['View'].insertSeparator
      @menuid['SGF']=@menu['View'].insertItem("SGF",self,"viewSGF")
      s=EUCstr.new('ΡɤSGF').toUTF
      @menuid['nodeSGF']=@menu['View'].insertItem(s,self,'nodeSGF')
      s=EUCstr.new('ƥȸ').toUTF
      @menuid['textGoban']=@menu['View'].insertItem(s,self,'textGoban')

      s=EUCstr.new('Ѳޤ(BS)').toUTF
      @menuid['returnFromBranch']=@menu['Move'].insertItem(s,self,"returnFromBranch")
      s=EUCstr.new('').toUTF
      @menuid['backToMain']=@menu['Move'].insertItem(s,self,"backToMain")
      @menu['Move'].insertSeparator
      s=EUCstr.new('ʬ(&F)').toUTF
      @menuid['gotoNextBranchPoint']=@menu['Move'].insertItem(s,self,"gotoNextBranchPoint")
      s=EUCstr.new('ľʬ(&B)').toUTF
      @menuid['gotoLastBranchPoint']=@menu['Move'].insertItem(s,self,"gotoLastBranchPoint")
      @menu['Move'].insertSeparator
      s=EUCstr.new('Υ(&N)').toUTF
      @menuid['nextComment']=@menu['Move'].insertItem(s,self,"nextComment")
      s=EUCstr.new('ľΥ(&P)').toUTF
      @menuid['lastComment']=@menu['Move'].insertItem(s,self,"lastComment")
      @menu['Move'].insertSeparator
      s=EUCstr.new('Ф').toUTF
      @menuid['goBackToStone']=@menu['Move'].insertItem(s,self,"goBackToStone")
      s=EUCstr.new('ֹ').toUTF
      @menuid['gotoNumber']=@menu['Move'].insertItem(s,self,"gotoNumber")

      @menu['moveMethod']=QPopupMenu.new
      s=EUCstr.new('̥å').toUTF
      @menuid['cursorOff']=@menu['moveMethod'].insertItem(s,self,"setCursorOff")
      s=EUCstr.new('Ȥ').toUTF
      @menuid['cursorOn']=@menu['moveMethod'].insertItem(s,self,"setCursorOn")
      s=EUCstr.new('').toUTF
      @menuid['modeMove']=@menu['Opt'].insertItem(s,@menu['moveMethod'])
      @menu['Opt'].insertSeparator

      s=EUCstr.new('Ȥɽʤ').toUTF
      @menuid['optComment']=@menu['Opt'].insertItem(s,self,'optComment')
      s=EUCstr.new('ɽʤ').toUTF
      @menuid['optMark']=@menu['Opt'].insertItem(s,self,'optMark')
      @menu['optNumber']=QPopupMenu.new
      s=EUCstr.new('ʤ').toUTF
      @menuid['optNumber1']=@menu['optNumber'].insertItem(s,self,"optNumber1")
      s=EUCstr.new('').toUTF
      @menuid['optNumber2']=@menu['optNumber'].insertItem(s,self,"optNumber2")
      s=EUCstr.new('Ǹ').toUTF
      @menuid['optNumber3']=@menu['optNumber'].insertItem(s,self,'optNumber3')
      s=EUCstr.new('ֹɽ').toUTF
      @menuid['optNumber']=@menu['Opt'].insertItem(s,@menu['optNumber'])
     @menu['optNumber'].setItemChecked(@menuid['optNumber3'],true)
   end

   def newTree(s=19)
      if @tree!=nil && @tree.edited?
         r=ConfirmDialog.new(self,1).exec
         return if r!=QMessageBox::Yes
      end
      @fileName=nil
      @tree=GameTree.new(s)
      @tree.analyze
      @goban.changeSize(@tree.sz)
      @undo.clear
      drawNode
      setCaption()
   end
   def newTree15
      newTree(15)
   end
   def newTree13
      newTree(13)
   end
   def newTree9
      newTree(9)
   end

   def load
      if @tree.edited?
         r=ConfirmDialog.new(self,1).exec
         return if r!=QMessageBox::Yes
      end
      fileOpen=FileDialog.new(self)
      r=fileOpen.exec
      return if r!=1
      if fileOpen.selectedFile==nil then
         fileOpen.myAccept
         return if fileOpen.selectedFile==nil
      end
      begin
         file=open(fileOpen.selectedFile)#Ȥ߹ߴؿ
         file.gets(nil)
         file.close
         Dir.chdir(File.dirname(fileOpen.selectedFile))
      rescue
         s=EUCstr.new('եɤ߹ߤ˼Ԥޤ').toUTF
         Warning.new(self,s).exec
         return
      end
      if $_==nil||$_[0..500].index('(') == nil
         Warning.new(self,'File does not seem SGF').exec
         return
      end
      sgf=SGF.new($_)
      sgf.gsub!(/\r\n|\r/e,"\n")

      tree=sgf.toGameTree
      if tree==nil
         s = EUCstr.new(sgf.errMsg).toUTF
         Warning.new(self,s).exec
         return
      end
      if sgf.garvage!=nil
         s  = EUCstr.new(sgf.garvage).toUTF
         QMessageBox.new('',s,QMessageBox::Information,1,0,0,self,'').exec
      end
      @tree=tree
      @goban.changeSize(@tree.sz)

      @fileName=fileOpen.selectedFile
      @fileWritable=FileTest.writable?(File.dirname(@fileName))
      @fileWritable=@fileWritable&&FileTest.writable?(@fileName)

      @undo.clear
      drawNode
      setCaption()
      return
   end
   def loadFromTextArea
      if @tree.edited?
         r=ConfirmDialog.new(self,1).exec
         return if r!=QMessageBox::Yes
      end
      sgf=SGF.new(@txtArea.text)
      tree=sgf.toGameTree
      if tree==nil
         s = EUCstr.new(sgf.errMsg).toUTF
         Warning.new(self,s).exec
         return
      end
      if sgf.garvage!=nil
         s  = EUCstr.new(sgf.garvage).toUTF
         QMessageBox.new('',s,QMessageBox::Information,1,0,0,self,'').exec
      end

      @tree=tree
      @tree.setEdited(true)
      @fileName=nil
      @goban.changeSize(@tree.sz)
      @undo.clear
      drawNode
      setCaption()
   end

   def save
      return if !@tree.edited?
      return if @fileName==nil
      return if !@fileWritable

      s=@fileName+"\n˾¸ޤ\nǤ"
      s=EUCstr.new(s).toUTF
      r=ConfirmDialog.new(self,s).exec
      return if r!=QMessageBox::Yes
      begin
         file=open(@fileName,'w')
         file.write(@tree.toSGF)
         file.close
      rescue
         s=EUCstr.new('ե¸˼Ԥޤ').toUTF
         Warning.new(self,s).exec
         return
      end
      @tree.setEdited(false)
      @menu['File'].setItemEnabled(@menuid['save'],false)
   end
   def saveAs
      fileOpen=FileDialog.new(self,'w')
      r=fileOpen.exec
      return if r!=1
      if fileOpen.selectedFile==nil then
         fileOpen.myAccept
         return if fileOpen.selectedFile==nil
      end
      begin
         file=open(fileOpen.selectedFile,'w')
         file.write(@tree.toSGF)
         file.close
         Dir.chdir(File.dirname(fileOpen.selectedFile))
         s=fileOpen.selectedFile+"\n¸ޤ"
         s=EUCstr.new(s).toUTF
         QMessageBox.new('',s,QMessageBox::Information,1,0,0,self,'').exec
      rescue
         s=EUCstr.new('ե¸˼Ԥޤ').toUTF
         Warning.new(self,s).exec
         return
      end
         
      @fileName=fileOpen.selectedFile
      @fileWritable=FileTest.writable?(File.dirname(@fileName))
      @fileWritable=@fileWritable&&FileTest.writable?(@fileName)

      @tree.setEdited(false)
      drawNode
      setCaption()
   end

   def boardPointedAt(xy)
      x,y=xy
      return if x<1 || y<1 || x>@tree.sz || y>@tree.sz

      if @status==Normal
         if @tree.currentNode.board[x,y]==Board::NoStone
            addNewMove(x,y)
            drawNode
         end
      elsif @status==GoBackToStone
         if @tree.currentNode.board[x,y]!=Board::NoStone
            @tree.goBackToStone(x,y)
            goBackToStoneEnd
            drawNode
         end
      elsif @status==InitialBoard
         c=@initialBoardView.color
         if @tree.currentNode.board[x,y]==Board::NoStone
            if c!=nil
               err=@tree.addInitialStone(c,x,y)
               if err==nil
                  @tree.gameinfo['HA']=nil
                  @tree.setEdited(true) 
               end
            end
         else
            @tree.removeInitialStone(x,y)
            @tree.gameinfo['HA']=nil
            @tree.setEdited(true)
         end
         drawNode(:dontChangeText)
      elsif @status==PutMark
         pn=@putMarkView.propName
         return if pn==nil
         pt=(?a+x-1).chr+(?a+y-1).chr
         PropList::PropMarkup.each{|propName|
            pv=@tree.currentNode.propList[propName]
            next if pv==nil
            pv.each_index{|i|
               if pv[i][0,2]==pt
                  pv[i]=nil
                  @tree.setEdited(true)
               end
            }
            pv.compact!
            if pv.empty?
               @tree.currentNode.propList.delete(propName)
            end
         }
         if pn!=''
            if pn=='LB'
               ch=@putMarkView.lbChr
               pt+= ':'+ch
            elsif pn=='LB2'
               ch=@putMarkView.lbChr2
               pn='LB'
               pt+= ':'+ch
               @putMarkView.lb2Succ
            end
            if @tree.currentNode.propList[pn]==nil
               pv=PropValue.new
               pv.push(pt)
               @tree.currentNode.addProp(pn,pv)
            else
               @tree.currentNode.propList[pn].push(pt)
            end
            @tree.setEdited(true)
         end
         drawNode(:dontChangeText)
      elsif @status==CountArea
         countAreaPoint(x,y)
      elsif @status==MoveStone
         moveStoneExec(x,y)
      end
   end

   def addNewMove(x,y)
      col=@tree.nextMoveColor
      i=@tree.newMove?(col,x,y)
      if i==true
         bPar=@tree.currentNode.board
         b=bPar.move(col,x,y)
         if b.error!=nil
            Warning.new(self,Board,b.error).exec
            return
         end
         if x==0
            pt='tt'
         else
            pt=(?a+x-1).chr+(?a+y-1).chr
         end
         node=Node.new
         node.move=Move.new(col,pt)
         node.setBoard(b)
         @undo.pushStack(@tree.saveStatus(:current))
         i=@tree.addChild(node)
      end
      @tree.gotoChild(i)
   end
   def pass
      addNewMove(0,0)
      drawNode
   end

   def undo
      if @undo.size>0
         @tree=@undo.undo(@tree)
         @tree.setEdited(true)
      end
      drawNode
   end
   def redo # redoͽ(湽¤)
      if @undo.redoAvailable?
         @tree=@undo.redo(@tree)
         drawNode
      end
   end

   def addComment
      s=@txtArea.text
      t=AddCommentDialog.new(self,'',s)
      r=t.exec
      if r==QDialog::Accepted
         commentOrg=''
         cn=@tree.currentNode
         if cn.propList['C']!=nil
            commentOrg=cn.propList['C'][0]
         end
         s=EUCstr.new(t.text)
         if s!=commentOrg
            @undo.pushStack(@tree.saveStatus(:current))
            cn=@tree.currentNode
            if cn.propList['C']!=nil
               cn.propList.delete('C') 
            end
            unless s.empty?
               pv=PropValue.new
               pv.push(s)
               cn.addProp('C',pv)
            end
            @tree.setEdited(true)
            drawNode
         end
      end
   end

   def removeNode
      @undo.pushStack(@tree.saveStatus(:parent))
      @tree.removeCurrentNode
      drawNode
   end

   def kakuage(mode)
      treecopy=@tree.saveStatus(:parent)
      r=@tree.kakuage(mode)
      if r!=nil
         @undo.pushStack(treecopy)
         drawNode
      end
   end
   def kakuageToA
      kakuage(GameTree::KakuageToA)
   end
   def kakuage1
      kakuage(GameTree::Kakuage1)
   end
   def kakusage1
      kakuage(GameTree::Kakusage1)
   end
   def kakusageToZ
      kakuage(GameTree::KakusageToZ)
   end


   def initialBoard
      @status=InitialBoard
      @treecopy=@tree.saveStatus(:all)

      if @initialBoardView==nil
         @initialBoardView=InitialBoardView.new
         connect(@initialBoardView.ok, QSIGNAL("clicked()"), self, 'initialBoardEnd')
         connect(@initialBoardView.allDelete, QSIGNAL("clicked()"), self, 'initialBoardAllDelete')
         connect(@initialBoardView.putOkiisiSignal, self, 'putOkiisi')
         idd=@initialBoardView.id
         @ctrlViewStack.addWidget(@initialBoardView,idd)
      end
      @initialBoardView.reset(@tree.sz)
      @ctrlViewStack.raiseWidget(@initialBoardView)

      @initialBoardView.setPL(@tree.propPL)
      @gobanView.setFocus
   end
   def initialBoardAllDelete
      f=@tree.removeInitialStoneAll
      @tree.setEdited(true) if f
      if @tree.gameinfo['HA']
         @tree.gameinfo['HA']=nil
         @tree.setEdited(true)
      end
      drawNode(:dontChangeText)
   end
   def initialBoardEnd
      plNew=@initialBoardView.pl
      if @tree.propPL!=plNew
         @tree.propPL=plNew
         @tree.setEdited(true)
      end
      if @tree.rootNode2SGF!=@treecopy.rootNode2SGF
         @undo.pushStack(@treecopy)
         @tree.deleteBoard
      end
      @treecopy=nil
      @ctrlViewStack.raiseWidget(@ctrlView) 
      @status=Normal

      drawNode(:dontChangeText)
   end

   def putOkiisi(okiisi)
      @tree.putOkiisi(okiisi)
      drawNode(:dontChangeText)
   end

   def moveStone
      @status=MoveStone
      if @moveStoneView==nil
         @moveStoneView=MoveStoneView.new
         connect(@moveStoneView.cancel, QSIGNAL("clicked()"), self, 'moveStoneEnd')
         idd=@moveStoneView.id
         @ctrlViewStack.addWidget(@moveStoneView,idd)
      end
      @ctrlViewStack.raiseWidget(@moveStoneView)
      @optionSave=@option.dup
      @option['mark']=false
      @option['number']=false
      drawNode(:dontChangeText)
      @goban.drawMoveStoneMark([@tree.currentNode.move.x,@tree.currentNode.move.y])
   end
   def moveStoneEnd
      @goban.drawMoveStoneMark(nil)
      @ctrlViewStack.raiseWidget(@ctrlView) 
      @status=Normal
      @option=@optionSave
      drawNode(:dontChangeText)
   end
   def moveStoneExec(x,y)
      cn=@tree.currentNode
      return if cn.board[x,y]!=Board::NoStone
      ip=cn.parent
      c =cn.move.color
      return if @tree[ip].board.move(c,x,y).error

      @undo.pushStack(@tree.saveStatus(:all))
      @tree.moveStone(x,y)
      moveStoneEnd
   end

   def putMark
      @status=PutMark
      @treecopy=@tree.saveStatus(:current)
      if @putMarkView==nil
         @putMarkView=PutMarkView.new
         connect(@putMarkView.ok, QSIGNAL("clicked()"), self, 'putMarkEnd')
         connect(@putMarkView.bAllDelete, QSIGNAL("clicked()"), self,'putMarkAllDelete' )
         idd=@putMarkView.id
         @ctrlViewStack.addWidget(@putMarkView,idd)
      else
         @putMarkView.reset
      end
      @ctrlViewStack.raiseWidget(@putMarkView)
   end
   def putMarkEnd
      if @tree.currentNode.toSGF!=@treecopy.currentNode.toSGF
         @undo.pushStack(@treecopy)
      end
      @treecopy=nil

      @ctrlViewStack.raiseWidget(@ctrlView) 
      @status=Normal
      drawNode(:dontChangeText)
   end
   def putMarkAllDelete
      PropList::PropMarkup.each{|pn|
         @tree.currentNode.propList.delete(pn)
      }
      @tree.setEdited(true)
      drawNode(:dontChangeText)
   end

   def current2Initial
      @undo.pushStack(@tree.saveStatus(:all))
      @tree.current2Initial
      @goban.clear
      drawNode
   end

   def slotAlphabetPressed(i)
      if i==?P-?A
         pass
      else
         gotoVariation(i) if @status==Normal
      end
   end

   def slotShiftAlphabetPressed(i)
      if @status==Normal
         case i
         when ?N 
            nextComment
         when ?P
            lastComment
         when ?F
            gotoNextBranchPoint
         when ?B
            gotoLastBranchPoint
         when ?C
            addComment
         when ?D
            removeNode
         when ?U
            undo
         when ?R
            self.redo # redoͽ(湽¤)
         end
      end
   end

   def slotBSpressed
     returnFromBranch if @status==Normal && @tree.inBranch?
   end
   def slotCursorPressed(i)
      return if @status!=Normal
      if !@tree.currentNode.child.empty?
         case i
         when 1
            nextMove
         when 10
            next10Move
         when 1000
            lastMove
         end
      end
      if @tree.currentNode.parent
         case i
         when -1
            prevMove
         when -10
            prev10Move
         when -1000
            initialMove
         end
      end
   end

   def slotTabPressed
      case @status
      when Normal
         if @gobanView.hasFocus()
            @txtArea.setFocus()
         else
            @gobanView.setFocus()
         end
      when InitialBoard
         if @initialBoardView.focusChange==nil
            @gobanView.setFocus()
         end
      when PutMark
         if @putMarkView.focusChange==nil
            @gobanView.setFocus()
         end
      end
   end

   def nextMove
      @tree.gotoChild(0)
      drawNode
   end
   def next10Move
      for i in 1..10
         break if @tree.gotoChild(0) == nil
      end
      drawNode
   end
   def lastMove
      i=0
      f=false
      while !@tree.currentNode.child.empty?
        i+=1
        if i==100
          i=0
          f=(@tree[@tree.currentNode.child[0]].board==nil)
        end
        @tree.gotoChild(0)
        if f
          drawNode
          f=false
        end
      end
      drawNode
   end
   def prevMove
      @tree.prev
      drawNode
   end
   def prev10Move
      for i in 1..10
         break if (@tree.prev) == nil
      end
      drawNode
   end
   def initialMove
      while @tree.prev != nil
      end
      drawNode
   end

   def gotoVariation(i)
      f=@tree.gotoVariation(i)
      drawNode if f
   end

   def goBackToStone
      @status=GoBackToStone
      if @goBackToStoneView==nil
         @goBackToStoneView=GoBackToStoneView.new
         connect(@goBackToStoneView.cancel, QSIGNAL("clicked()"), self, 'goBackToStoneEnd')
         idd=@goBackToStoneView.id
         @ctrlViewStack.addWidget(@goBackToStoneView,idd)
      end
      @ctrlViewStack.raiseWidget(@goBackToStoneView)
   end
   def goBackToStoneEnd
      @status=Normal
      @ctrlViewStack.raiseWidget(@ctrlView) 
      drawNode(:dontChangeText)
   end

   def returnFromBranch
      @tree.returnFromBranch
      drawNode
   end
   def backToMain
      while @tree.inBranch?
         @tree.returnFromBranch
      end
      drawNode
   end

   def gotoLastBranchPoint
      return nil if !@tree.lastBranchPoint?
      @tree.gotoLastBranchPoint
      drawNode
   end
   def gotoNextBranchPoint
      return nil if @tree.nextBranchPoint==nil
      @tree.gotoNextBranchPoint
      drawNode
   end
 
   def lastComment
      return nil if !@tree.lastComment?
      @tree.gotoLastComment
      drawNode
   end
   def nextComment
      return nil if @tree.nextComment==nil
      @tree.gotoNextComment
      drawNode
   end
   def gotoNumber
      max=@tree.lastNumber
      n=@tree.currentNode.board.n
      d=GotoNumberDialog.new(self,n,max)
      r=d.exec
      if r==1
         @tree.gotoNumber(d.value)
         drawNode
      end
      d=nil
   end

   def setCursorOn
     @menu['moveMethod'].setItemChecked(@menuid['cursorOff'],false)
     @menu['moveMethod'].setItemChecked(@menuid['cursorOn'],true)
     @goban.setCursorOn(true)
   end
   def setCursorOff
     @menu['moveMethod'].setItemChecked(@menuid['cursorOff'],true)
     @menu['moveMethod'].setItemChecked(@menuid['cursorOn'],false)
     @goban.setCursorOn(false)
   end

   def optComment
     f=!@menu['Opt'].isItemChecked(@menuid['optComment'])
     @menu['Opt'].setItemChecked(@menuid['optComment'],f)
     @option['comment']=!f
     drawNode
   end

   def optMark
     f=!@menu['Opt'].isItemChecked(@menuid['optMark'])
     @menu['Opt'].setItemChecked(@menuid['optMark'],f)
     @menu['Edit'].setItemEnabled(@menuid['putMark'],!f)
     @option['mark']=!f
     drawNode(:dontChangeText)
   end

   def optNumber1
     @menu['optNumber'].setItemChecked(@menuid['optNumber1'],true)
     @menu['optNumber'].setItemChecked(@menuid['optNumber2'],false)
     @menu['optNumber'].setItemChecked(@menuid['optNumber3'],false)
     @option['number']=false
     drawNode(:dontChangeText)
   end
   def optNumber2
     @menu['optNumber'].setItemChecked(@menuid['optNumber1'],false)
     @menu['optNumber'].setItemChecked(@menuid['optNumber2'],true)
     @menu['optNumber'].setItemChecked(@menuid['optNumber3'],false)
     @option['number']=true
     drawNode(:dontChangeText)
   end
   def optNumber3
     @menu['optNumber'].setItemChecked(@menuid['optNumber1'],false)
     @menu['optNumber'].setItemChecked(@menuid['optNumber2'],false)
     @menu['optNumber'].setItemChecked(@menuid['optNumber3'],true)
     @option['number']=3
     drawNode(:dontChangeText)
   end

   def viewHama
      b=@tree.currentNode.board
      s  = sprintf("äС%3d\n",b.hamaB)
      s += sprintf("򤬼äС%3d\n",b.hamaW)
      s=EUCstr.new(s).toUTF
      q=QMessageBox.new('',s,QMessageBox::NoIcon,1,0,0,self,'')
      q.move(10,10)
      q.exec
      q=nil
      GC.start
   end
   def viewGameinfo
      GameinfoDialog.new(@tree.gameinfo,'v',self,'').exec
   end
   def editGameinfo
      d= GameinfoDialog.new(@tree.gameinfo,'e',self,'')
      c=d.exec
      if c==1
         edited=false
         PropList::PropGameinfo.each{|p|
            next if d.lineEdit[p]==nil
            if d.lineEdit[p].edited
               unless edited
                  edited=true
                  @treecopy=@tree.saveStatus(:root)
               end
               s=EUCstr.new(d.lineEdit[p].text)
               if @tree.gameinfo[p]==nil
                  @tree.gameinfo[p]=PropValue.new
               end
               @tree.gameinfo[p][0]=s
               @tree.setEdited(true)
            end
         }
         if edited
            @undo.pushStack(@treecopy)
            @treecopy=nil
         end
         drawNode(:dontChangeText)
      end
   end
   def viewSGF
      qstr=@tree.toSGF.toUTF
      @txtArea.setText(qstr)
   end
   def nodeSGF
      if @tree.currentNode.parent==nil
         qstr=EUCstr.new(@tree.rootNode2SGF).toUTF
      else
         qstr=EUCstr.new(@tree.currentNode.toSGF).toUTF
      end
      @txtArea.setText(qstr)
   end
   def textGoban
      s=@tree.currentNode.board.toStr
      s=EUCstr.new(s).toUTF
      TextGobanDialog.new(self,'',s).exec
   end

   def drawNode(changeText=true)
      while @tree.currentNode.board==nil
         break if @tree.currentNode.parent==nil
         @tree.prev
      end

      @moveView.draw(@tree.currentNode)

      cn=@tree.currentNode
      if cn.board.error!=nil
         Warning.new(self,Board,cn.board.error).exec
      end
      @goban.draw(cn,@option)
      @gobanView.show

      if changeText==true
         str=''
         if @option['comment']
            c=cn.propList['C']
            if c!=nil
               str=c[0]
            end
         end
         qstr=EUCstr.new(str).toUTF
         @txtArea.setText(qstr)
      end

      @buttonBox.draw(cn)
      @variationView.draw(@tree)

      @menu['File'].setItemEnabled(@menuid['save'],(@fileName!=nil && @fileWritable && @tree.edited?))

      f=@tree.currentNode.move.is_a?(Move)
      @menu['Edit'].setItemEnabled(@menuid['removeNode'],f)
      f=f&&(@tree.currentNode.move.x!=0)
      @menu['Edit'].setItemEnabled(@menuid['moveStone'],f)

      f=false
      iPar=@tree.currentNode.parent
      if iPar!=nil
         nodeP=@tree[iPar]
         if nodeP.child.size>1
            m=@menu['kakuage']
            fTmp=(@tree.currentIndex!=nodeP.child[0])
            m.setItemEnabled(@menuid['kakuageToA'], fTmp)
            m.setItemEnabled(@menuid['kakuage1'],   fTmp)
            fTmp=(@tree.currentIndex!=nodeP.child.last)
            m.setItemEnabled(@menuid['kakusageToZ'],fTmp)
            m.setItemEnabled(@menuid['kakusage1'],  fTmp)
            f=true
         end
      end
      @menu['Edit'].setItemEnabled(@menuid['kakuage'],f)

      @menu['Edit'].setItemEnabled(@menuid['initialBoard'],(@tree.currentNode.parent==nil))
      @menu['Edit'].setItemEnabled(@menuid['current2Initial'],(@tree.currentNode.parent!=nil))

      @menu['Edit'].setItemEnabled(@menuid['undo'],!@undo.empty?)
      @menu['Edit'].setItemEnabled(@menuid['redo'],@undo.redoAvailable?)

      if @tree.currentNode.parent==nil || @tree[@tree.currentNode.parent].parent==nil
         @menu['Move'].setItemEnabled(@menuid['goBackToStone'],false)
      else
         @menu['Move'].setItemEnabled(@menuid['goBackToStone'],true)
      end
      @menu['Move'].setItemEnabled(@menuid['gotoNumber'],(@tree.currentNode.parent!=nil || @tree.currentNode.child.size>0))

      @menu['Move'].setItemEnabled(@menuid['returnFromBranch'],@tree.inBranch?)
      @menu['Move'].setItemEnabled(@menuid['backToMain'],@tree.inBranch?)
      @menu['Move'].setItemEnabled(@menuid['gotoNextBranchPoint'],(@tree.nextBranchPoint!=nil))
      @menu['Move'].setItemEnabled(@menuid['gotoLastBranchPoint'],@tree.lastBranchPoint?)
      @menu['Move'].setItemEnabled(@menuid['nextComment'],(@tree.nextComment!=nil))
      @menu['Move'].setItemEnabled(@menuid['lastComment'],@tree.lastComment?)
      @gobanView.setFocus()
   end

   def countArea
      @status=CountArea
      if @countAreaView==nil
         @countAreaView=CountAreaView.new
         connect(@countAreaView.ok, QSIGNAL("clicked()"), self, 'countAreaEnd')
         idd=@countAreaView.id
         @ctrlViewStack.addWidget(@countAreaView,idd)
         @listOfdeadStone=Array.new
      end
      @ctrlViewStack.raiseWidget(@countAreaView)

      @optionSave=@option.dup
      @option['mark']=false
      @option['number']=false
      drawNode(:dontChangeText)
      countAreaCount
   end
   def countAreaPoint(x,y)
      if @tree.currentNode.board[x,y]!=Board::NoStone
         ind=@tree.currentNode.board.index(x,y)
         ren=@tree.currentNode.board.ren(ind)
         if @listOfdeadStone.include?([x,y])
            ren.each{|ii|
               xx,yy=@tree.currentNode.board.i2ij(ii)
               @listOfdeadStone.delete([xx,yy])
            }
         else
            ren.each{|ii|
               xx,yy=@tree.currentNode.board.i2ij(ii)
               @listOfdeadStone.push([xx,yy])
            }
         end
         @goban.drawDeadstone(@listOfdeadStone)
         countAreaCount
      end
   end
   def countAreaCount
      s=@tree.currentNode.board.countArea(@listOfdeadStone)
      @countAreaView.txtArea.setText(s)
   end
   def countAreaEnd
      @countAreaView.reset
      @ctrlViewStack.raiseWidget(@ctrlView) 
      @status=Normal

      @listOfdeadStone.clear
      @goban.drawDeadstone(@listOfdeadStone)
      @option=@optionSave
      drawNode(:dontChangeText)
   end
   def slotTextChange
      @menu['File'].setItemEnabled(@menuid['loadFromTextArea'],@txtArea.isSGF?)
   end
   def closeEvent(ev)
      if @status!=Normal
         ev.ignore
         if @status==GoBackToStone
            goBackToStoneEnd
         end
         if @status==MoveStone
            moveStoneEnd
         end
         return
      end
      if @tree.edited?
         r=ConfirmDialog.new(self,ConfirmDialog::Quit).exec
         if r!=QMessageBox::Yes
            ev.ignore
            return
         end
      end
      p "closeEvent"
      print "PWD=",Dir.pwd,"\n"
      ev.accept
   end
end
