class Node

   def initialize
      @propList  = PropList.new
      @child = Array.new 
      @board = nil
      @move  = nil
      @parent= nil
   end
   attr_reader   :propList
   attr_accessor :parent
   attr_accessor :child
   attr_reader   :board
   attr_accessor :move

   def setBoard(b)
      @board=b
   end

   def addProp(id,val)
      if id=='B'||id=='W'
         if @move==nil
            v, = val # v=val[0]
            if v=~/^[a-z][a-z]$/||v.empty?
               @move=Move.new(id,v)
            end
         end
      elsif id!='CA' && id!='AP'
         @propList.addProp(id,val)
      end
   end

   def addChild(i)
      @child.push(i)
   end

   def toSGF
      s=';'
      s<<@move.toSGF if @move
      s<<propList.toSGF
      return s
   end

   def semiDeepCopy
      tmp=@board
      @board=nil
      c=Marshal.load(Marshal.dump(self))
      @board=tmp
      c.setBoard(tmp)
      c
   end
end

class PropList < Hash
   PropRoot = ['FF','GM','SZ']
   PropGameinfo=['PB','PW','BR','WR','DT','EV','PC','RE',\
                 'HA','KM','SO']
   PropSetup= ['AB','AW','PL']
   PropMarkup=['LB','MA','TR','CR','SQ','SL']
   def toSGF
      s=''
      nch=0
      self.each_key{|id|
         next if PropRoot.include?(id)
         next if PropSetup.include?(id)
         next if PropGameinfo.include?(id)
         s<<id
         nch+=id.size
         str=self[id].toSGF
         s<<str
         nch+=str.size
         if nch>=SGF::LineLength
            s<<"\n"
            nch=0
         end
      }
      return s
   end

   def addProp(id,val)
      if !val.is_a?(PropValue)
         print "Err in PropList.addProp: val is not PropValue\n"
         return nil
      end
      self[id]=val
      return true
   end
end

class PropValue < Array
   def toSGF
      s=''
      self.each{|val|
         s<<'['
         val=val.gsub(/\\/e,'\\\\\\')
         val=val.gsub(/\]/e,'\\\]')
         s<<val
         s<<']'
      }      
      return s
   end
end

class Move
   attr_reader :x
   attr_reader :y
   attr_reader :color
   def initialize(color,pt)
      @color=color
      @pt=pt
      if pt.empty?
         x=y=0
      else
         x=pt[0]-?a+1
         y=pt[1]-?a+1
      end
      if x>=1&&y>=1&&x<=52&&y<=52
         @x=x
         @y=y
      else
         @x=@y=0
      end
   end

   def pass
       @x=@y=0
   end

   def toSGF
      return color+'['+@pt+']'
   end
end
